@@ -32,7 +32,7 @@ var stateMigrationQuery string
 
 type commentRow = struct {
 	CommenthistoryId uint64
-	EntryTime        int64
+	EntryTime        sql.NullInt64
 	EntryTimeUsec    uint32
 	EntryType        uint8
 	AuthorName       string
@@ -50,12 +50,16 @@ type commentRow = struct {
 func convertCommentRows(
 	env string, envId icingadbTypes.Binary,
 	_ func(interface{}, string, ...interface{}), _ *sqlx.Tx, idoRows []commentRow,
-) (icingaDbInserts, _ [][]contracts.Entity, checkpoint any) {
+) (stages []icingaDbOutputStage, checkpoint any) {
 	var commentHistory, acknowledgementHistory, allHistoryComment, allHistoryAck []contracts.Entity
 
 	for _, row := range idoRows {
 		checkpoint = row.CommenthistoryId
 
+		if !row.EntryTime.Valid {
+			continue
+		}
+
 		typ := objectTypes[row.ObjecttypeId]
 		hostId := calcObjectId(env, row.Name1)
 		serviceId := calcServiceId(env, row.Name1, row.Name2)
@@ -63,7 +67,7 @@ func convertCommentRows(
 		switch row.EntryType {
 		case 1: // user
 			id := calcObjectId(env, row.Name)
-			entryTime := convertTime(row.EntryTime, row.EntryTimeUsec)
+			entryTime := convertTime(row.EntryTime.Int64, row.EntryTimeUsec)
 			removeTime := convertTime(row.DeletionTime, row.DeletionTimeUsec)
 			expireTime := convertTime(row.ExpirationTime, 0)
 
@@ -129,7 +133,7 @@ func convertCommentRows(
 				name += "!" + row.Name2
 			}
 
-			setTime := convertTime(row.EntryTime, row.EntryTimeUsec)
+			setTime := convertTime(row.EntryTime.Int64, row.EntryTimeUsec)
 			setTs := float64(setTime.Time().UnixMilli())
 			clearTime := convertTime(row.DeletionTime, row.DeletionTimeUsec)
 			acknowledgementHistoryId := hashAny([]any{env, name, setTs})
@@ -146,19 +150,9 @@ func convertCommentRows(
 				},
 				AckHistoryUpserter: history.AckHistoryUpserter{ClearTime: clearTime},
 				SetTime:            setTime,
-				Author: icingadbTypes.String{
-					NullString: sql.NullString{
-						String: row.AuthorName,
-						Valid:  true,
-					},
-				},
-				Comment: icingadbTypes.String{
-					NullString: sql.NullString{
-						String: row.CommentData,
-						Valid:  true,
-					},
-				},
-				ExpireTime: convertTime(row.ExpirationTime, 0),
+				Author:             icingadbTypes.MakeString(row.AuthorName),
+				Comment:            icingadbTypes.MakeString(row.CommentData),
+				ExpireTime:         convertTime(row.ExpirationTime, 0),
 				IsPersistent: icingadbTypes.Bool{
 					Bool:  row.IsPersistent != 0,
 					Valid: true,
@@ -207,7 +201,12 @@ func convertCommentRows(
 		}
 	}
 
-	icingaDbInserts = [][]contracts.Entity{commentHistory, acknowledgementHistory, allHistoryComment, allHistoryAck}
+	stages = []icingaDbOutputStage{
+		{insert: commentHistory},
+		{insert: acknowledgementHistory},
+		{insert: allHistoryComment},
+		{insert: allHistoryAck},
+	}
 	return
 }
 
@@ -218,8 +217,9 @@ type downtimeRow = struct {
 	CommentData         string
 	IsFixed             uint8
 	Duration            int64
-	ScheduledStartTime  int64
+	ScheduledStartTime  sql.NullInt64
 	ScheduledEndTime    int64
+	WasStarted          uint8
 	ActualStartTime     int64
 	ActualStartTimeUsec uint32
 	ActualEndTime       int64
@@ -236,17 +236,21 @@ type downtimeRow = struct {
 func convertDowntimeRows(
 	env string, envId icingadbTypes.Binary,
 	_ func(interface{}, string, ...interface{}), _ *sqlx.Tx, idoRows []downtimeRow,
-) (icingaDbInserts, _ [][]contracts.Entity, checkpoint any) {
+) (stages []icingaDbOutputStage, checkpoint any) {
 	var downtimeHistory, allHistory, sla []contracts.Entity
 
 	for _, row := range idoRows {
 		checkpoint = row.DowntimehistoryId
 
+		if !row.ScheduledStartTime.Valid || row.WasStarted == 0 {
+			continue
+		}
+
 		id := calcObjectId(env, row.Name)
 		typ := objectTypes[row.ObjecttypeId]
 		hostId := calcObjectId(env, row.Name1)
 		serviceId := calcServiceId(env, row.Name1, row.Name2)
-		scheduledStart := convertTime(row.ScheduledStartTime, 0)
+		scheduledStart := convertTime(row.ScheduledStartTime.Int64, 0)
 		scheduledEnd := convertTime(row.ScheduledEndTime, 0)
 		triggerTime := convertTime(row.TriggerTime, 0)
 		actualStart := convertTime(row.ActualStartTime, row.ActualStartTimeUsec)
@@ -357,13 +361,17 @@ func convertDowntimeRows(
 		sla = append(sla, s)
 	}
 
-	icingaDbInserts = [][]contracts.Entity{downtimeHistory, allHistory, sla}
+	stages = []icingaDbOutputStage{
+		{insert: downtimeHistory},
+		{insert: allHistory},
+		{insert: sla},
+	}
 	return
 }
 
 type flappingRow = struct {
 	FlappinghistoryId  uint64
-	EventTime          int64
+	EventTime          sql.NullInt64
 	EventTimeUsec      uint32
 	EventType          uint16
 	PercentStateChange sql.NullFloat64
@@ -377,7 +385,7 @@ type flappingRow = struct {
 func convertFlappingRows(
 	env string, envId icingadbTypes.Binary,
 	selectCache func(dest interface{}, query string, args ...interface{}), _ *sqlx.Tx, idoRows []flappingRow,
-) (icingaDbInserts, icingaDbUpserts [][]contracts.Entity, checkpoint any) {
+) (stages []icingaDbOutputStage, checkpoint any) {
 	if len(idoRows) < 1 {
 		return
 	}
@@ -402,7 +410,11 @@ func convertFlappingRows(
 	for _, row := range idoRows {
 		checkpoint = row.FlappinghistoryId
 
-		ts := convertTime(row.EventTime, row.EventTimeUsec)
+		if !row.EventTime.Valid {
+			continue
+		}
+
+		ts := convertTime(row.EventTime.Int64, row.EventTimeUsec)
 
 		// Needed for ID (see below).
 		var start icingadbTypes.UnixMilli
@@ -506,15 +518,18 @@ func convertFlappingRows(
 		}
 	}
 
-	icingaDbInserts = [][]contracts.Entity{flappingHistory, allHistory}
-	icingaDbUpserts = [][]contracts.Entity{flappingHistoryUpserts}
+	stages = []icingaDbOutputStage{
+		{insert: flappingHistory},
+		{upsert: flappingHistoryUpserts},
+		{insert: allHistory},
+	}
 	return
 }
 
 type notificationRow = struct {
 	NotificationId     uint64
 	NotificationReason uint8
-	EndTime            int64
+	EndTime            sql.NullInt64
 	EndTimeUsec        uint32
 	State              uint8
 	Output             string
@@ -528,7 +543,7 @@ type notificationRow = struct {
 func convertNotificationRows(
 	env string, envId icingadbTypes.Binary,
 	selectCache func(dest interface{}, query string, args ...interface{}), ido *sqlx.Tx, idoRows []notificationRow,
-) (icingaDbInserts, _ [][]contracts.Entity, checkpoint any) {
+) (stages []icingaDbOutputStage, checkpoint any) {
 	if len(idoRows) < 1 {
 		return
 	}
@@ -579,6 +594,10 @@ func convertNotificationRows(
 	for _, row := range idoRows {
 		checkpoint = row.NotificationId
 
+		if !row.EndTime.Valid {
+			continue
+		}
+
 		previousHardState, ok := cachedById[row.NotificationId]
 		if !ok {
 			continue
@@ -597,7 +616,7 @@ func convertNotificationRows(
 			continue
 		}
 
-		ts := convertTime(row.EndTime, row.EndTimeUsec)
+		ts := convertTime(row.EndTime.Int64, row.EndTimeUsec)
 		tsMilli := float64(ts.Time().UnixMilli())
 		notificationHistoryId := hashAny([]interface{}{env, name, ntEnum, tsMilli})
 		id := hashAny([]interface{}{env, "notification", name, ntEnum, tsMilli})
@@ -627,13 +646,8 @@ func convertNotificationRows(
 			SendTime:          ts,
 			State:             row.State,
 			PreviousHardState: previousHardState,
-			Text: icingadbTypes.String{
-				NullString: sql.NullString{
-					String: text,
-					Valid:  true,
-				},
-			},
-			UsersNotified: row.ContactsNotified,
+			Text:              icingadbTypes.MakeString(text),
+			UsersNotified:     row.ContactsNotified,
 		})
 
 		allHistory = append(allHistory, &history.HistoryNotification{
@@ -665,7 +679,11 @@ func convertNotificationRows(
 		}
 	}
 
-	icingaDbInserts = [][]contracts.Entity{notificationHistory, userNotificationHistory, allHistory}
+	stages = []icingaDbOutputStage{
+		{insert: notificationHistory},
+		{insert: userNotificationHistory},
+		{insert: allHistory},
+	}
 	return
 }
 
@@ -702,12 +720,12 @@ func convertNotificationType(notificationReason, state uint8) icingadbTypes.Noti
 
 type stateRow = struct {
 	StatehistoryId      uint64
-	StateTime           int64
+	StateTime           sql.NullInt64
 	StateTimeUsec       uint32
 	State               uint8
 	StateType           uint8
-	CurrentCheckAttempt uint16
-	MaxCheckAttempts    uint16
+	CurrentCheckAttempt uint32
+	MaxCheckAttempts    uint32
 	LastState           uint8
 	LastHardState       uint8
 	Output              sql.NullString
@@ -721,7 +739,7 @@ type stateRow = struct {
 func convertStateRows(
 	env string, envId icingadbTypes.Binary,
 	selectCache func(dest interface{}, query string, args ...interface{}), _ *sqlx.Tx, idoRows []stateRow,
-) (icingaDbInserts, _ [][]contracts.Entity, checkpoint any) {
+) (stages []icingaDbOutputStage, checkpoint any) {
 	if len(idoRows) < 1 {
 		return
 	}
@@ -744,13 +762,17 @@ func convertStateRows(
 	for _, row := range idoRows {
 		checkpoint = row.StatehistoryId
 
+		if !row.StateTime.Valid {
+			continue
+		}
+
 		previousHardState, ok := cachedById[row.StatehistoryId]
 		if !ok {
 			continue
 		}
 
 		name := strings.Join([]string{row.Name1, row.Name2}, "!")
-		ts := convertTime(row.StateTime, row.StateTimeUsec)
+		ts := convertTime(row.StateTime.Int64, row.StateTimeUsec)
 		tsMilli := float64(ts.Time().UnixMilli())
 		stateHistoryId := hashAny([]interface{}{env, name, tsMilli})
 		id := hashAny([]interface{}{env, "state_change", name, tsMilli})
@@ -776,10 +798,10 @@ func convertStateRows(
 			HardState:         row.LastHardState,
 			PreviousSoftState: row.LastState,
 			PreviousHardState: previousHardState,
-			CheckAttempt:      uint8(row.CurrentCheckAttempt),
+			CheckAttempt:      row.CurrentCheckAttempt,
 			Output:            icingadbTypes.String{NullString: row.Output},
 			LongOutput:        icingadbTypes.String{NullString: row.LongOutput},
-			MaxCheckAttempts:  uint32(row.MaxCheckAttempts),
+			MaxCheckAttempts:  row.MaxCheckAttempts,
 			CheckSource:       icingadbTypes.String{NullString: row.CheckSource},
 		})
 
@@ -819,6 +841,10 @@ func convertStateRows(
 		}
 	}
 
-	icingaDbInserts = [][]contracts.Entity{stateHistory, allHistory, sla}
+	stages = []icingaDbOutputStage{
+		{insert: stateHistory},
+		{insert: allHistory},
+		{insert: sla},
+	}
 	return
 }
